package explorviz.hpc_monitoring.filter.reduction;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.RingBuffer;
import com.lmax.disruptor.dsl.Disruptor;

import explorviz.hpc_monitoring.disruptor.RecordArrayEvent;
import explorviz.hpc_monitoring.disruptor.RecordEvent;
import explorviz.hpc_monitoring.filter.counting.CountingThroughputFilter;
import explorviz.hpc_monitoring.reader.TimeProvider;
import explorviz.hpc_monitoring.record.IRecord;
import explorviz.hpc_monitoring.record.TerminateRecord;
import explorviz.hpc_monitoring.record.TimedPeriodRecord;
import explorviz.hpc_monitoring.record.trace.Trace;

public class TracePatternSummarizationFilter implements EventHandler<RecordArrayEvent> {
	private static final int RINGBUFFER_LENGTH = 256;

	private final long maxCollectionDuration;

	private final Map<Trace, TraceAggregationBuffer> trace2buffer = new ConcurrentSkipListMap<Trace, TraceAggregationBuffer>(
			new TraceComperator());

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Reduced traces/sec");

	private final RingBuffer<RecordEvent> ringBuffer;

	public TracePatternSummarizationFilter(final long maxCollectionDuration,
			final EventHandler<RecordEvent> endReceiver) {
		this.maxCollectionDuration = maxCollectionDuration;

		final ExecutorService exec = Executors.newCachedThreadPool();
		final Disruptor<RecordEvent> disruptor = new Disruptor<RecordEvent>(
				RecordEvent.EVENT_FACTORY, RINGBUFFER_LENGTH, exec);

		@SuppressWarnings("unchecked")
		final EventHandler<RecordEvent>[] eventHandlers = new EventHandler[1];
		eventHandlers[0] = endReceiver;
		if (endReceiver != null) {
			disruptor.handleEventsWith(eventHandlers);
		}
		ringBuffer = disruptor.start();
	}

	@Override
	public void onEvent(final RecordArrayEvent event, final long sequence, final boolean endOfBatch)
			throws Exception {
		final IRecord[] values = event.getValues();
		final int valuesLength = event.getValuesLength();

		for (int i = 0; i < valuesLength; i++) {
			final IRecord record = values[i];
			if (record instanceof Trace) {
				final Trace trace = (Trace) record;
				if (trace.isValid()) {
					insertIntoBuffer(trace);
				} else {
					sendOutTrace(trace);
				}
			} else if (record instanceof TimedPeriodRecord) {
				processTimeoutQueue(TimeProvider.getCurrentTimestamp());
			} else if (record instanceof TerminateRecord) {
				terminate();
			}
		}
	}

	private void insertIntoBuffer(final Trace trace) {
		TraceAggregationBuffer traceAggregationBuffer = trace2buffer.get(trace);
		if (traceAggregationBuffer == null) {
			traceAggregationBuffer = new TraceAggregationBuffer(TimeProvider.getCurrentTimestamp());
			trace2buffer.put(trace, traceAggregationBuffer);
		}
		traceAggregationBuffer.insertTrace(trace);
	}

	private void processTimeoutQueue(final long timestamp) {
		final long bufferTimeout = timestamp - maxCollectionDuration;
		final List<Trace> toRemove = new ArrayList<Trace>();
		for (final TraceAggregationBuffer traceBuffer : trace2buffer.values()) {
			if (traceBuffer.getBufferCreatedTimestamp() <= bufferTimeout) {
				final Trace aggregatedTrace = traceBuffer.getAggregatedTrace();
				sendOutTrace(aggregatedTrace);
				toRemove.add(aggregatedTrace);
			}
		}
		for (final Trace traceEventRecords : toRemove) {
			trace2buffer.remove(traceEventRecords);
		}
	}

	private void sendOutTrace(final Trace aggregatedTrace) {
		counter.inputObjects(aggregatedTrace);
		final long hiseq = ringBuffer.next();
		final RecordEvent valueEvent = ringBuffer.get(hiseq);
		valueEvent.setValue(aggregatedTrace);
		ringBuffer.publish(hiseq);
	}

	private void terminate() {
		for (final TraceAggregationBuffer traceBuffer : trace2buffer.values()) {
			sendOutTrace(traceBuffer.getAggregatedTrace());
		}
		trace2buffer.clear();
	}
}
